library immutable_map;

import 'package:option/option.dart';
import 'package:immutable_list/immutable_list.dart';
import 'dart:convert';

class ImmutableMap<K, V> {

  final Map<K, V> _map;

  /**
   * Const constructor for const initialization
   */
  const ImmutableMap(): _map = const {};

  /**
   * named constructor for initializing from a mutable map.
   */
  ImmutableMap.fromMap(Map<K, V> map): _map = new Map.from(map);

  /**
   * Checks if the map contains a key.
   *
   * Returns boolean true if the key is contained in the map, false
   * otherwise.
   */
  bool containsKey(K key) => _map.containsKey(key);

  /**
   * Checks if the map contains a value.
   *
   * Returns boolean true if the value is contained in the map, false
   * otherwise.
   */
  bool containsValue(V value) => _map.containsValue(value);

  /**
   * Calls the supplied function on each key value pair.
   */
  void forEach(void f(K, V)) => _map.forEach(f);

  /**
   * Creates an new map where the key is set.
   *
   * If the value is already set in the map then this just returns itself.
   * Otherwise, it returns a new map where the key is set to the result of
   * calling the supplied ifAbsent method.
   */
  ImmutableMap<K, V> putIfAbsent(K key, V ifAbsent()) {
    final val = _map[key];
    return val == null
      ? set(key, ifAbsent())
      : this;
  }

  /**
   * Optionally returns the value associated to the key.
   *
   * If the key exists then it's value is returned wrapped in a Some<V>, a
   * None<V> otherwise.
   */
  Option<V> get(K key) => new Option(_map[key]);

  /**
   * Creates a new map with the supplied key set.
   *
   * Returns a new map with the supplied key set to the supplied value.
   * If the key already existed it's value will be overwritten.
   */
  ImmutableMap<K, V> set(K key, V value) {
    var map = new Map.from(_map);
    map[key] = value;
    return new ImmutableMap.fromMap(map);
  }

  /**
   * Creates a new map without the supplied key.
   *
   * Returns a new map that doesn't contain the supplied key.
   */
  ImmutableMap<K, V> remove(K key) {
    var map = new Map.from(_map);
    map.remove(key);
    return new ImmutableMap.fromMap(map);
  }

  /**
   * Returns an immutable list of all the keys in this map.
   */
  ImmutableList<K> keys() => new ImmutableList.fromIterable(_map.keys);

  /**
   * Returns an immutable list of all the values in this map.
   */
  ImmutableList<V> values() => new ImmutableList.fromIterable(_map.values);

  /**
   * Merges two maps together returning a new map as the result.
   */
  ImmutableMap<K, V> merge(ImmutableMap<K, V> other) {
    var map = new Map.from(_map);
    map.addAll(other._map);
    return new ImmutableMap.fromMap(map);
  }

  /**
   * Returns the json string encoded representation.
   */
  String toJson() => new JsonEncoder().convert(_map);

}